@@ -10,4 +10,8 @@ $self->{LD}="CC -ztext";
 $self->{CCCDLFLAGS}="-KPIC";
 $self->{clean} = {FILES => 'Templates.DB'};
 $self->{PERLMAINCC} = 'gcc';
-$self->{LIBS} = ["-R$ENV{OS_ROOTDIR}/lib -loscol -los -losths -lC"];
+
+use Config;
+my $thrlib = $Config{usethreads} ? '-losthr' : '-losths';
+$self->{LIBS} = ["-R$ENV{OS_ROOTDIR}/lib -loscol -los $thrlib -lC"];
+#$self->{LIBS} = ["-R$ENV{OS_ROOTDIR}/lib.sym -loscol -los $thrlib -lC"];
